
import { Engine, Gui } from "./export_interface";

export interface FramCallCbInsertType {
    direction: InsertDirection;
    anchor: () => void;
}

export enum InsertDirection {
    Before,
    After,
}

export class FramCall {
    static reset_render_state: boolean;
    
    static addFramCallBefore(callback: () => void): void;
    static addFramCall(callback: () => void): void;
    static addFramCallAfter(callback: () => void): void;
    
    static removeFramCallBefore(f: () => void): void;
    static removeFramCall(f: () => void): void;
    static removeFramCallAfter(f: () => void): void;

    static addFramCallBeforeInterval(callback: () => void, insertType?: FramCallCbInsertType): void;
    static addFramCallInterval(callback: () => void): void;
    static addLastFramCallPostInterval(callback: () => void): void;
    static addFramCallAfterInterval(callback: () => void, insertType?: FramCallCbInsertType): void;

    static removeFramCallBeforeInterval(f: () => void): void;
    static removeFramCallInterval(f: () => void): void;
    static removeFramCallAfterInterval(f: () => void): void;
}

// 开始帧推
export const startFramCall: () => void;

export interface DocumentOption {
    bottom?: number;
    right?: number;
    top?: number;
    left?: number;
    viewPortWidth?: number;
    viewPortHeight?: number;
    isRotate?: boolean;
    styleWidth?: number;
    styleHeight?: number;
    pixilRato?: number;
    class_sheet?: number;
    font_sheet?: number;
}

export interface OptionEngineD3 {
    particlesystem: boolean;
    skeleton: boolean;
    shadowmapping: boolean;
    lighting: boolean;
    spine: boolean;
}

export let engine: Engine;
export let gui_obj: Engine;

export class GuiContext {
    width: number;
    height: number;
    isRotate: boolean;
    viewPortWidth: number;
    viewPortHeight: number;
    styleWidth: number;
    styleHeight: number;
    pixelRatio: number;
    scaleX: number;
    scaleY: number;
    radioX: number;
    radioY: number;
    viewPort: {
    x: number;
    y: number;
    width: number;
    height: number;
    };
    uniqueID: Gui;
    root: number;
    engine: Engine;

    constructor(
    width: number,
    height: number,
    renderEngine: HTMLCanvasElement,
    asset_allotor: any,
    asset_total_capacity: number,
    asset_config: string,
    engine_log_filter: string,
    engine_log_level: number,
    gui_mod: number,
    _use_sdf: boolean,
    option?: DocumentOption
    );

    initSize(width: number, height: number, option?: DocumentOption): void;
    setView(): void;
    static createRoot(context: Gui, width: number, height: number): number;
}