export { atomTable, AtomTable } from "./utils";
import { atomTable } from "./utils";
import { Atom } from "./dom";


/**
 * 获取atom的number， 用于将字符串数字化
 * 一些静态字符串，底层不关心字符串的具体值，通常是选择一个数字进行高层和底层的通信
 * @param v 
 * @returns 
 */
export function getAtomnumber(v: number | string | Atom): number {
	if (typeof v === "number") {
		return v;
	} else if (typeof v === "string") {
		let r = atomTable.get(v);
		if (!r) {
			let atom = Atom.from_string(v);
			let number = atom.get_hash()
			// number = murmurhash3_32s(v, 0);
			atomTable.set(number, v, atom);
			return number;
		}
		return r[0];
	} else {
		// Atom
		let number = v.get_hash();
		return number;
	}
}

export function getString(v: number | string | Atom): string | undefined{
	if (typeof v === "string") {
		return v;
	} else if (typeof v === "number") {
		let r = atomTable.get(v);
		if (!r) {
			let s = Atom.get_string_by_hash(v);
			if (s !== undefined && s !== null) {
				let atom = Atom.from_string(s);
				atomTable.set(v, s, atom);
			} else {
				throw new Error("No string found！！！， key:" + v);
			}
			return s;
		}
		return r[1];
	} else {
		let r = atomTable.get(v.get_hash());
		if (r) {
			return r[1];
		}
	}
}

export function getAtom(v: number | string | Atom): Atom {
	if (v instanceof Atom) {
		return v;
	}
	let r = atomTable.get(v);
	if (r) {
		if (r[2] == undefined) {
			r[2] = Atom.from_string(r[1]);
		}
		return r[2];
	} else if (typeof v == "string") {
		let atom = Atom.from_string(v);
		let number = atom.get_hash()
		// number = murmurhash3_32s(v, 0);
		atomTable.set(number, v, atom);
		return atom;
	} else {
		throw new Error("No atom found！！！， key:" + v);
	}
}

export function get(v: number | string): [number, string, Atom] {
	return atomTable.get(v)
}