/**
 * 在本地平台上， 需要用gui绘制加载进度条等， 但此时不希望加载gui的全部代码， 本模块实现一个简单版本的dom，供前期调用
 */
import * as engine_interface from "./export_interface";
import {getGlobal} from "pi_ucore/util/frame_mgr";
import * as atom from "./atom";
import { PISYS } from 'pi_sys/index';
// import * as word_load from "pi_common/word_load/word_load";
 
import {setAtomFunc} from "./snippets/gui_web-fe567f277be011de/js/utils"; 
setAtomFunc(atom.getString);

/// 帧回调插入类型
export interface FramCallCbInsertType {
	direction: InsertDirection,
	anchor: () => void
}

export enum InsertDirection {
	Before, // 插入到某回调之前
	After,	// 插入到某回调之后
}

/**
 * 开始帧推
 */
export const startFramCall = () => {
	getGlobal().setPermanentBefore(FramCall.framCall); // 帧推
}

interface FramCallBack {cur: Function[], next: Function[]}

export class FramCall {
	// 是否在当前帧渲染之前重置渲染状态
	public static reset_render_state: boolean;
	//每帧会清空
	private static _framCallsBefore: FramCallBack = {cur: [], next: []};
	//每帧会清空
	private static _framCalls: FramCallBack = {cur: [], next: []};
	//每帧会清空
	private static _framCallsAfter: FramCallBack = {cur: [], next: []};

	//每帧不会清空
	private static _framCallsBeforeInterval: Function[] = [];
	//每帧不会清空
	private static _framCallsInterval: Function[] = [];
	//每帧不会清空
	private static _framCallsAfterInterval: Function[] = [];
	//每帧不会清空
	private static _lastFramCallsPostInterval: Function[] = [];


	public static addFramCallBefore(callback: () => void) {
		FramCall._framCallsBefore.cur.push(callback);
	}

	public static addFramCall(callback: () => void) {
		FramCall._framCalls.cur.push(callback);
	}

	public static addFramCallAfter(callback: () => void) {
		FramCall._framCallsAfter.cur.push(callback);
	}

	public static removeFramCallBefore(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCallsBefore.cur );
	}

	public static removeFramCall(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCalls.cur );
	}

	public static removeFramCallAfter(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCallsAfter.cur );
	}

	public static addFramCallBeforeInterval(callback: () => void, insertType?: FramCallCbInsertType) {
		if (!insertType) {
			FramCall._framCallsBeforeInterval.push(callback);
		} else {
			let i = FramCall._framCallsBeforeInterval.indexOf(insertType.anchor);
			if (i == -1) {
				throw new Error("FramCall.addFramCallBeforeInterval: anchor not found!");
			}

			if (insertType.direction === InsertDirection.Before) {
				FramCall._framCallsBeforeInterval.splice(i, 0, callback);
			} else {
				FramCall._framCallsBeforeInterval.splice(i + 1, 0, callback);
			}
		}
	}

	public static addFramCallInterval(callback: () => void) {
		FramCall._framCallsInterval.push(callback);
	}

	public static addLastFramCallPostInterval(callback: () => void) {
		FramCall._lastFramCallsPostInterval.push(callback);
	}

	public static addFramCallAfterInterval(callback: () => void, insertType?: FramCallCbInsertType) {
		if (!insertType) {
			FramCall._framCallsAfterInterval.push(callback);
		} else {
			let i = FramCall._framCallsAfterInterval.indexOf(insertType.anchor);
			if (i == -1) {
				throw new Error("FramCall.addFramCallAfterInterval: anchor not found!");
			}

			if (insertType.direction === InsertDirection.Before) {
				FramCall._framCallsAfterInterval.splice(i, 0, callback);
			} else {
				FramCall._framCallsAfterInterval.splice(i + 1, 0, callback);
			}
		}
	}

	public static removeFramCallBeforeInterval(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCallsBeforeInterval );
	}

	public static removeFramCallInterval(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCallsInterval );
	}

	public static removeFramCallAfterInterval(f: () => void) {
		FramCall.removeFramCallFunc(f, FramCall._framCallsInterval );
	}


	public static framCall() {
		FramCall.call(FramCall._lastFramCallsPostInterval);
		FramCall.framCallBCack(FramCall._framCallsBefore);
		FramCall.call(FramCall._framCallsBeforeInterval);
		if (engine) {
			FramCall.framCallBCack(FramCall._framCalls);
			FramCall.call(FramCall._framCallsInterval);
			engine_interface.fram_call(engine, !!FramCall.reset_render_state);
		}
		FramCall.framCallBCack(FramCall._framCallsAfter);
		FramCall.call(FramCall._framCallsAfterInterval);
	}

	private static framCallBCack(_framCalls: FramCallBack) {
		let temp = _framCalls.cur;
		_framCalls.cur = _framCalls.next;
		_framCalls.next = temp;
		FramCall.call(_framCalls.next);
		_framCalls.next.length = 0;
	}
	

	private static call(list: Function[]) {
		// 调用回调
		if (list.length > 0) {
			for (var c of list) {
				c();
			}
		}
	}

	private static removeFramCallFunc(f: () => void, framCall: Function[]) {
		if (!framCall) {
			return;
		}
		let len = framCall.length;
		for (let i = 0; i < len; i++) {
			if (framCall[i] === f) {
				framCall.splice(i, 1);
			}
		}
	}
	

}

export interface DocumentOption {

	/**
	 * 相对于屏幕在x轴小的偏移（如果渲染到屏幕上）
	 */
	bottom?: number;
	/**
	 * 相对于屏幕在Y轴又的偏移（如果渲染到屏幕上）
	 */
	right?: number;
	/**
	 * 相对于屏幕在x轴小的偏移（如果渲染到屏幕上）
	 */
	top?: number;
	/**
	 * 相对于屏幕在Y轴又的偏移（如果渲染到屏幕上）
	 */
	left?: number;
	/**
	 * 视口宽度
	 */
	viewPortWidth?: number;
	/**
	 * 视口高度
	 */
	viewPortHeight?: number;
	/**
	 * 如果是true， 将顺时针旋转渲染内容，并旋转渲染目标
	 */
	isRotate?: boolean;

	/**
	 * css宽度
	 */
	styleWidth?: number;
	/**
	 * css高度
	 */
	styleHeight?: number;

	/**
	 * 像素比率
	 */
	pixilRato?: number;

	class_sheet?: number; // 共用的样式表

	font_sheet?: number; // 共用的字体表

	// createGUIResNew?: Function; // 创建纹理资源回调， 只有web版本会用到
}

export interface OptionEngineD3 {
	particlesystem: boolean,
	skeleton: boolean,
	shadowmapping: boolean,
	lighting: boolean,
	spine: boolean,
}

export let engine: engine_interface.Engine;
export let gui_obj: engine_interface.Engine;

export class GuiContext {
	width: number;
	height: number;
	isRotate: boolean;
	viewPortWidth: number;
	viewPortHeight: number;
	styleWidth: number;
	styleHeight: number;
	pixelRatio: number;
	scaleX: number;
	scaleY: number;
	radioX: number;
	radioY: number;
	viewPort: {
		x: number,
		y: number,
		width: number,
		height: number,
	};

	uniqueID: engine_interface.Gui;
	root: number;
	engine: engine_interface.Engine;
	/**
	 * 创建Document
	 * @param width 根节点宽度(布局使用的宽度)
	 * @param height 根节点高度(布局使用的高度)
	 * @param renderEngine 渲染上下文， 当需要重用其他gui中的资源时使用RenderEngine
	 */
	constructor(
		width: number, 
		height: number, 
		renderEngine: HTMLCanvasElement,
		asset_allotor: any,
		asset_total_capacity: number,
		asset_config: string,
		engine_log_filter: string,
		engine_log_level: number,
		gui_mod: number,
		_use_sdf: boolean,
		option?: DocumentOption
	) {
		// option?: DocumentOption
		
		// this.index = Document.documents.length;
		// Document.documents.push(this);

		console.log("create gui=========");

		// let time = performance.now();
		this.initSize(width, height, option);

		let engine3dparam = new Uint32Array(32);
		let optionD3 = PISYS.Env.get("optionEngineD3") || {
			active: true,
			particlesystem: true,
			skeleton: true,
			shadowmapping: true,
			lighting: true,
			spine: true,
			disenable_ubo: false,
			disenable_material_array: false,
			maxlen_material_array: 256,
			max_instance_batch_count: 1024 * 1024,
			max_texture_size: 2048
		};
		engine3dparam[ 0] = optionD3.active ? 1 : 0;
		engine3dparam[ 1] = optionD3.particlesystem ? 1 : 0;
		engine3dparam[ 2] = optionD3.lighting ? 1 : 0;
		engine3dparam[ 3] = optionD3.shadowmapping ? 1 : 0;
		engine3dparam[ 4] = optionD3.skeleton ? 1 : 0;
		engine3dparam[ 5] = optionD3.directshadowmapping ? 1 : 0;
		engine3dparam[ 6] = optionD3.pointshadowmapping ? 1 : 0;
		engine3dparam[ 7] = optionD3.spotshadowmapping ? 1 : 0;
		engine3dparam[ 8] = optionD3.combinebuffersize || 1024 * 1024;
		engine3dparam[ 9] = optionD3.disenable_ubo == false ? 0 : 1;
		engine3dparam[10] = optionD3.disenable_material_array == false ? 0 : 1;
		engine3dparam[11] = optionD3.maxlen_material_array || 256;
		engine3dparam[12] = optionD3.max_instance_batch_count || 1024 * 1024;
		engine3dparam[13] = optionD3.max_texture_size || 2048;
		let optionSpine = PISYS.Env.get("optionSpine") || { active: true };
		if (!engine) {
			console.log("create_engine start1=========", );
			// 创建渲染引擎
			engine = engine_interface.create_engine(renderEngine, this.viewPortWidth, this.viewPortHeight, asset_allotor, asset_total_capacity || 600 * 1024 * 1024, asset_config || "", engine_log_filter || "", engine_log_level || 3, 4000);
			if (optionD3.active) {
				engine_interface.init_engine_3d(engine, optionD3.spine, engine3dparam);
			}
			if (optionSpine.active) {
				engine_interface.init_spine_context(engine);
			}
			window["engine"] = engine;
			console.log("create_engine end=========", engine, window["eeeeeee"] ===  engine);
		}
		this.engine = engine;

		// 创建gui引擎
		if (!gui_obj) {
			let load_sdf = undefined;
			// if (use_sdf) {
			// 	load_sdf = (key, name, chars: Uint32Array) => {
			// 		// let font_name = atom.getString(name);
			// 		// (new word_load.WordLoad()).startLoad(font_name, chars).then((buffers) => {
			// 		// 	engine_interface.sdf_on_load(key, buffers);
			// 		// });
			// 	};
			// }
			
			gui_obj = engine_interface.create_gui(this, engine, width, height, load_sdf, option.class_sheet ? option.class_sheet : 0, option.font_sheet ? option.font_sheet : 0, 0, 0, gui_mod || 0);
			FramCall.addFramCallInterval(() => {
				engine_interface.render_gui(gui_obj, engine);
			});
		}
		this.uniqueID = gui_obj;
		

		if (gui_mod !== 2) {
			// 创建根节点
			this.root = GuiContext.createRoot(this.uniqueID, width, height);

			// 设置视口
			this.setView();
		}
	}

	public static createRoot(context: engine_interface.Gui, width: number, height: number) {
		// 创建根节点
		let root = engine_interface.create_node(context);

		// 设置 GUI 清屏颜色
		// gui.set_clear_color(this.uniqueID, 0, 0, 0, 0, this.first.uniqueID, true);
		engine_interface.set_width(context, root, width);
		engine_interface.set_height(context, root, height);
		engine_interface.set_position_type(context, root, 1); // absolute
		engine_interface.set_position(context, root, 0, 0); // left
		engine_interface.set_position(context, root, 1, 0); // top
		engine_interface.set_margin(context, root, 0, 0); // left
		engine_interface.set_margin(context, root, 1, 0); // top

		// 将root作为根节点（如果不设置为根节点，该节点及其子节点不会参与渲染）
		engine_interface.insert_as_root(context, root);

		return root;

	}

	/**
	 * 改变尺寸
	 * @param width 根节点宽度(布局使用的宽度)
	 * @param height 根节点高度(布局使用的高度)
	 */
	public initSize(width: number, height: number, option?: DocumentOption) {
		// super({document: null, uniqueID: 1 });
		option = option || {};
		this.width = width;
		this.height = height;
		// this.offsetX = option.offsetX?option.offsetX:0;
		// this.offsetY = option.offsetY?option.offsetY:0;
		// this.bottom = option.bottom?option.bottom:0;
		// this.top = option.top?option.top:0;
		// this.left = option.left?option.left:0;
		// this.right = option.right?option.right:0;
		this.isRotate = option.isRotate ? option.isRotate : false;
		this.viewPortWidth = option.viewPortWidth ? option.viewPortWidth : width;
		this.viewPortHeight = option.viewPortHeight ? option.viewPortHeight : height;
		this.styleWidth = option.styleWidth ? option.styleWidth : width;
		this.styleHeight = option.styleHeight ? option.styleHeight : height;
		this.pixelRatio = option.pixilRato ? option.pixilRato : 1;

		this.scaleX = this.styleWidth / this.width;
		this.scaleY = this.styleHeight / this.height;
		this.radioX = this.viewPortWidth / this.width;
		this.radioY = this.viewPortHeight / this.height;

		this.viewPort = {
			x: 0,
			y: 0,
			width: this.viewPortWidth,
			height: this.viewPortHeight,
		};
	}

	// 设置视口
	public setView() {
		engine_interface.set_view_port(this.uniqueID, 0, 0, this.viewPort.width, this.viewPort.height, this.root);


		// 顺时针旋转90度（底层接口， 逆时针为正）
		if (this.isRotate) {
			engine_interface.set_project_transfrom(this.uniqueID, 1 / this.radioX, 1 / this.radioY, 1 - this.radioX, this.radioY - 1, -90, this.width, this.height);
			// gui.set_project_transfrom(this.uniqueID, 1, 1, 0, 0, -90);
			engine_interface.set_view_port(this.uniqueID, this.viewPort.x, this.viewPort.y, this.viewPort.height, this.viewPort.width, this.root);
			engine_interface.set_scissor(this.uniqueID, 0, 0, this.viewPortHeight, this.viewPortWidth);
		} else {
			engine_interface.set_project_transfrom(this.uniqueID, 1 / this.radioX, 1 / this.radioY, 1 - this.radioX, this.radioY - 1, 0, this.width, this.height);
			engine_interface.set_view_port(this.uniqueID, this.viewPort.x, this.viewPort.y, this.viewPort.width, this.viewPort.height, this.root);
			engine_interface.set_scissor(this.uniqueID, 0, 0, this.viewPortWidth, this.viewPortHeight);
		}
		// 当布局尺寸不等于视口尺寸时，需要设置缩放，以保证字体清晰
		// if (this.radioX !== 1) {
		//     this.radioX = this.viewPort.width / this.width;
		//     gui.transform_origin(this.uniqueID, 1, 0, 0, 0, 0);
		//     gui.transform_scale(this.uniqueID, 1, this.radioX, this.radioY);
		// }
		console.log("yyyyyyyyyyyyyyyyyyyyyyyy");
		if (this.radioX !== 1) {
			console.log("xxxxxxxxxxxxxxxxxxxxxxx");
			this.radioX = this.viewPort.width / this.width;
			engine_interface.set_transform_origin(this.uniqueID, this.root, 0, 0, 0, 0);
			engine_interface.set_scale(this.uniqueID, this.root, `${this.radioX} ${this.radioY}`);
		}
	}
}

// pi_app
// export * from "./native/dom_init"
