import { PISYS } from "pi_sys/index"
import { requireSync } from "pi_sys/util/mod";
export * from "../../gui_web-fe567f277be011de/js/utils";
import { AtomTable, atomTable } from "../../gui_web-fe567f277be011de/js/utils";
import { loadRes } from "pi_sys/load/res_load";
import { Depend } from "pi_sys/util/depend";
import { request, create, initRequest } from 'pi_worker/client/client';
import { KTX } from 'pi_res/ktx';
import { Env } from "pi_sys/util/env";
import { appendFile, readdir, stat, readFile, access, mkdir, ftruncate } from "pi_sys/store/file_system";
import { traversalDir } from "pi_file_system_util/index";
// import { OutlineInfo, FontFace, Polyline, Path, Polygon, Ellipse, Segment, Rect, Circle } from "web/pi_sdf.wasm";

declare var wx;
const FontDirName = "_$fonts";
const getFontRootDir = () => {
    return `${wx.env.USER_DATA_PATH}/${FontDirName}`;
}

const getUnicodeFragmentStart = (unicode: number) => {
    if (unicode >= 0x0000 && unicode <= 0x2E7F) {
        return 0x0000;
    } else if (unicode >= 0x2E80 && unicode <= 0x4DFF) {
        return 0x2E80;
    } else if (unicode >= 0x4E00 && unicode <= 0x57C3) {
        return 0x4E00;
    } else if (unicode >= 0x57C4 && unicode <= 0x6187) {
        return 0x57C4;
    } else if (unicode >= 0x6188 && unicode <= 0x6B4B) {
        return 0x6188;
    } else if (unicode >= 0x6B4C && unicode <= 0x750F) {
        return 0x6B4C;
    } else if (unicode >= 0x7510 && unicode <= 0x7ED3) {
        return 0x7510;
    } else if (unicode >= 0x7ED4 && unicode <= 0x8897) {
        return 0x7ED4;
    } else if (unicode >= 0x8898 && unicode <= 0x925B) {
        return 0x8898;
    } else if (unicode >= 0x925C && unicode <= 0x9C1F) {
        return 0x825C;
    } else if (unicode >= 0x9C20 && unicode <= 0x9FA5) {
        return 0x9C20;
    } else if (unicode >= 0x9FA6 && unicode <= 0x323AF) {
        return 0x9FA6;
    } else {
        throw new Error(`not support unicode ${unicode}`);
    }
}

export function getGlobalMetricsHeight(font_name, size) {
    return size;
}

export function hasAtom(path: number) {
    return atomTable.get(path) != undefined;
}

export let setAtom = function (id: number, path: string) {
    // path = atomTable.get_string(path);
    // atomTable.set(id, path);
    // return PISYS.ResLoadTool.loadFile(path, undefined);
    // addAtomCall(atomTable, id, path);

    atomTable.set(id, path, undefined);
}

export function loadFile(path) {
    path = atomTable.get_string(path);
    return PISYS.ResLoadTool.loadFile(path, undefined);
}

export function loadImage(path) {
    path = atomTable.get_string(path);
    return PISYS.ResLoadTool.loadImage(path, undefined);
}

export function loadKtx(path) {
    path = atomTable.get_string(path);
    return PISYS.ResLoadTool.loadFile(path, undefined).then((r) => {
        let info = KTX.parse(r, false);
        let data = new Array(info.mipmaps.length);
        for (let i = 0; i < info.mipmaps.length; i++) {
            data[i] = info.mipmaps[i].data;
        }
        //  [width, height, depth, format, minmap_count, layer_count, face_count, [Data...]]
        return [info.width, info.height, 1, info.format, info.mipmapCount, 1, info.isCubemap ? 6 : 1, data]; //TODO depth、arrayCount都为1
    });
}

// 加载图片，并将图片draw在canvas上，返回canvas的2d上下文
export function loadImageAsCanvas(path) {
    let r = atomTable.get(path);
    if (!r) {
        return new Promise((resolve) => {
            let r = requireSync("pi_wasm/atom").getString(path);
            return loadImageAsCanvasInner(r).then((r) => {
                resolve(r)
            })
        });
    } else {
        return loadImageAsCanvasInner(r[1])
    }
}

function loadImageAsCanvasInner(path) {
    return PISYS.ResLoadTool.loadImage(path, undefined).then((image) => {
        let canvas = document.createElement("canvas");
        canvas.width = image.width;
        canvas.height = image.height;
        let ctx = canvas.getContext("2d");
        ctx.drawImage(image as HTMLImageElement, 0, 0);
        return ctx;
    })
}

enum PlatformType {
    WEB = "web",
    MINIGAME = "minigame",
    WEBVIEW = "webview",
    APP_WEBVIEW = "app_webview",
    ANDROID = "android",
    IOS = "ios",
    WINDOWS = "windows"
}

const platformType = PISYS.Env.get("platform_type") as PlatformType;

let workerCount = 1;
if (platformType === PlatformType.MINIGAME) {
    workerCount = 1;
}

let groupName = "fontSdf";

// ===================================================== 导出
let localStore: Store;
let initSuccess = false;
const dbName: string = "font";
const tabName: string = "sdf2";
let modName = "pi_wasm/web/worker";
let initWorkerSuccess = false;
let paths = [];


/**
 * 初始化indexDb
 */
let fontMap = new Map();
export const initLocalStore = (): Promise<void> => {
    const t1 = Date.now();
    // 初始化worker
    console.log("======== init worker  1, workerCount: ", workerCount, groupName);
    return create(groupName, { workerCount, wasmFilePaths: ["pi_wasm/web/pi_sdf.wasm"] }).then(() => {
        // 手动让worker加载wasm
        return request(groupName, modName, "initWasm", [], null, 1, 0).then(r => {
            // console.log("worker 执行 computeNearArcs  success: time:  ", performance.now() - begin);
            initWorkerSuccess = true; 
            console.log("======== init worker success 2");
            if (platformType === PlatformType.MINIGAME) {
                let rootDir = getFontRootDir();
                return access(rootDir).then(isExist => {
                    console.log("======== init worker success 3, ", rootDir);
                    if (!isExist) {
                        return Promise.resolve();
                    }
                    return traversalDir(rootDir, (path: string) => {
                        paths.push(path);
                    }).then(() => {
                        console.log("======== init worker success 4");
                        for (let path of paths) {
                            return readFile(path, "binary").then((data) => {
                                const arr = path.split("/");
                                const fontId = arr[arr.length - 2];
                                if (data && data.length > 0) {
                                    let offset = 0;
                                    while (offset < data.byteLength) {
                                        let view = new DataView(data.buffer, offset, 4);
                                        let code = view.getUint16(0);
                                        let len = view.getUint16(2);
            
                                        let buf = new Uint8Array(data.buffer, offset + 4, len);
                                        // console.log("init ", fontId, "; ", code, "; byteLength: ", len, buf.length, offset, data.buffer);
            
                                        let key = `${code}${fontId}`;
                                        fontMap.set(key, buf)
            
                                        offset += (4 + len);
                                    }
                                }
                            })
                        }
                    }).catch(e => {
                        console.error("======== init worker failed 5 ", e);
                    })
                })
            } else {
                return PISYS.Store.Store.createStore(dbName, tabName, 1).then(((store) => {
                    localStore = store;
                    initSuccess = true;
                    console.log("初始化字体存储完成", Date.now() - t1);
    
                }), ((reason) => {
                    console.error(`初始化字体存储错误, ${reason}`);
                }));
            }
        }).catch(e => {
            console.error("加载 sdfwasm  fail ", e);
        })
    }).catch(e => {
        console.error("======== init worker failed ", e);
    });
    // // console.log("======== init worker  2");
    // return Promise.resolve().then(()=>{
    //     console.log("======== init worker  3", Date.now(), Date.now() - t1);
    // });
};

/**
 * 从indexDb读数据
 */
// tslint:disable-next-line:no-reserved-keywords

export const get = async (key: string): Promise<any> => {
    // 微信小游戏，批处理
    if (platformType === PlatformType.MINIGAME) {
        let char_code = key.charCodeAt(0);
        // console.log("get: ", key, "; ", char_code);
        let font_name = key.slice(1, key.length);
        let map_key = `${char_code}${font_name}`;
        let buf = fontMap.get(map_key);
        if (!!buf) {
            fontMap.delete(map_key);
            // console.log("缓存有1:", key, "; ", char_code)
            return buf;
        } else {
            return null;
        }
    } else {
        // web 浏览器
        if (!initSuccess) return Promise.resolve();
        try {
            let data = await localStore.read(key);
            return data;
        } catch (e) {
            throw Error(e)
        }
    }
};

/**
 * 往indexDb写数据
 */
let set = new Set();
export const write = async (key: string, data: any) => {
    if (platformType === PlatformType.MINIGAME) {
        let char_code = key.charCodeAt(0);
        // console.log("write ", key, "; ", char_code);
        let font_name = key.slice(1, key.length);
        let code = getUnicodeFragmentStart(char_code);

        let rootDir = getFontRootDir();
        const dir = `${rootDir}/${font_name}`;
        const isExist = await access(dir);
        if (!isExist) {
            if (set.has(dir)) {
                return;
            }
            set.add(dir);
            await mkdir(dir);
        }
        const buffer = new ArrayBuffer(4);
        const dataView = new DataView(buffer);
        dataView.setInt16(0, char_code);
        dataView.setInt16(2, data.byteLength);
        const path = `${rootDir}/${font_name}/${code}`;

        let uint8View = new Uint8Array(new ArrayBuffer(4 + data.byteLength));
        uint8View.set(new Uint8Array(buffer), 0);
        uint8View.set(data, 4);

        // return appendFile(path, uint8View.buffer);
        return appendFile(path, uint8View.buffer).catch(e => {
            /**
             * appendFile失败,进行截断操作
             */
            console.error(`appendFile ${path} failed!!! len: ${4 + data.byteLength}, reason: ${e}`)
            ftruncate(path, 4 + data.byteLength).catch(e => {
                console.error(`pi_load ftruncate ${path} length = ${4 + data.byteLength} error`, e)
            });
        });
    } else {
        if (!initSuccess) return Promise.resolve();
        return localStore.write(key, data);
    }
};

/**
 * 从indexDb删除数据
 */
export const deleteKey = (key: string) => {
    if (platformType === PlatformType.MINIGAME) {
        return Promise.resolve();
    } else {
        if (!initSuccess) return;
        return localStore.delete(key);
    }
};

let FontFace, Circle, Rect, Segment, Ellipse, Polygon, Polyline, Path, SvgInfo, brotli_decompressor, OutlineInfo;

export function initSdfEnv(a, b, c, d, e, f, g, h, i, j, k) {
    FontFace = a;
    Circle = b;
    Rect = c;
    Segment = d;
    Ellipse = e;
    Polygon = f;
    Polyline = g;
    Path = h;
    SvgInfo = i;
    brotli_decompressor = j;
    OutlineInfo = k;
}

export const createFace = (font_data: Uint8Array) => {
    console.log("=============== font_data: ", font_data.length)
    return FontFace.new(font_data);
}



export const computeNearArcs = async (outline: any, scale: number) => {
    // let begin = performance.now();
    // let blob = new Blob([outline.buf]);
    // let url  = URL.createObjectURL(blob);
    if (initWorkerSuccess) {
        return request(groupName, modName, "computeNearArcs", [outline.buf.buffer, scale], null, 1, 0).then(r => {
            // console.log("worker 执行 computeNearArcs  success: time:  ", performance.now() - begin);
            return new Uint8Array(r);
        }).catch(e => {
            console.error("worker 执行 computeNearArcs fail ", e);
        })
    }else{
        let r = OutlineInfo.compute_near_arcs_of_wasm(outline.buf, scale);
        // console.log("同步执行 computeNearArcs  success: time:  ", performance.now() - begin);
        return r;
    }
}

export const computeSvgSdfTexOfWasm = async (info: any, tex_size: number, pxrange: number, is_outer_glow: boolean, cur_off: number, scale: number) => {
    let begin = performance.now();
    if (initWorkerSuccess) {
        return request(groupName, modName, "computeSvgSdfTexOfWasm", [info.buffer, tex_size, pxrange, is_outer_glow, cur_off, scale], null, 1, 0).then(r => {
            console.log("worker 执行 computeSvgSdfTexOfWasm  success: time:  ", performance.now() - begin);
            return new Uint8Array(r);
        }).catch(e => {
            console.error("worker 执行 computeSvgSdfTexOfWasm fail ", e);
        })
    } else {
        let r = SvgInfo.compute_sdf_tex_of_wasm(info, tex_size, pxrange, is_outer_glow, cur_off, scale);
         console.log("同步执行 computeSvgSdfTexOfWasm success: time:  ", performance.now() - begin);
        return r;
    }
}

export const computeSvgSdfTexOfWasmSync = (info: any, tex_size: number, pxrange: number, is_outer_glow: boolean, cur_off: number, scale: number) => {
    let begin = performance.now();
    let r = SvgInfo.compute_sdf_tex_of_wasm(info, tex_size, pxrange, is_outer_glow, cur_off, scale);
    console.log("同步执行 computeSvgSdfTexOfWasm success: time:  ", performance.now() - begin);
    return r;
}

export const computeLayout = (outline: any, tex_size: number, pxrange: number, cur_off: number,) => {
    let r = OutlineInfo.compute_layout_of_wasm(outline.extents, outline.units_per_em, tex_size, pxrange, cur_off);
    // console.log("compute_layout: ", r);
    return r;
}

export const computeSvgLayout = (binding_box: any, tex_size: number, pxrange: number, cur_off: number,) => {
    let r = SvgInfo.compute_layout_of_wasm(binding_box, tex_size, pxrange, cur_off);
    // console.log("compute_layout: ", r);
    return r;
}

export const computeSdfTexOfWasm = (outline: any, result_arcs: Uint8Array, tex_size: number, pxrange: number, is_outer_glow: boolean, cur_off: number,) => {
    // let begin = performance.now();
    let r = OutlineInfo.compute_sdf_tex_of_wasm(result_arcs, outline.extents, outline.units_per_em, outline.advance, tex_size, pxrange, is_outer_glow, cur_off);
    // console.log(" computeSdfTexOfWasm  success: time:  ", performance.now() - begin);
    return r;
}

export const computePositionsAndUv = (info:any, ps:Float32Array, uvs:Float32Array, half_stroke:number, ) => {
    let r = SvgInfo.compute_positions_and_uv_of_wasm(info, ps, uvs, half_stroke);
    // console.log("compute_layout: ", r);
    return r;
}

export const horizontalAdvance = (face: FontFace, text: string) => {
    return face.horizontal_advance(text);
}

export const horizontalAdvanceOfGlyphIndex = (face: FontFace, glyph_index: number) => {
    return face.horizontal_advance_of_glyph_index(glyph_index);
}

export const ascender = (face: FontFace) => {
    return face.ascender();
}

export const descender = (face: FontFace) => {
    return face.descender();
}

export const maxBox = (face: FontFace) => {
    return face.max_box();
}

export const maxBoxNormaliz = (face: FontFace) => {
    return face.max_box_normaliz();
}

export const toOutline = (face: FontFace, text: string) => {
    return face.to_outline_of_wasm(text);
}

export const toOutlineOfGlyphIndex = (face: FontFace, glyph_index: number) => {
    return face.to_outline_of_wasm_glyph_index(glyph_index);
}

export const glyphIndex = (face: FontFace, text: string) => {
    return face.glyph_index(text);
}

export const glyphIndexs = (face: FontFace, text: string, script: number) => {
    return face.glyph_indexs(text, script);
}

export const debugSize = (face: FontFace) => {
    return face.debug_size();
}

export const createCircle = (cx: number, cy: number, radius: number) => {
    return Circle.new(cx, cy, radius);
}

export const createRect = (x: number, y: number, width: number, height: number) => {
    return Rect.new(x, y, width, height);
}

export const createSegment = (ax: number, ay: number, bx: number, by: number) => {
    return Segment.new(ax, ay, bx, by);
}

export const createEllipse = (cx: number, cy: number, rx: number, ry: number) => {
    return Ellipse.new(cx, cy, rx, ry);
}

export const createPolygon = (points: Float32Array) => {
    return Polygon.new(points);
}

export const createPolyline = (points: Float32Array) => {
    return Polyline.new(points);
}

export const createPath = (verb: Uint8Array, points: Float32Array) => {
    return Path.new(verb, points);
}

export const getSvgInfo = (shape: any) => {
    let r = shape.get_svg_info_of_wasm();
    return r;
}

export const createSvgInfo = (bbox: Float32Array, arc_endpoints: Float32Array, is_area: boolean, is_reverse: boolean | undefined) => {
    let r = SvgInfo.new_of_wasm(bbox, arc_endpoints, is_area, is_reverse);
    return r;
}

export const free = (obj: any) => {
    if (obj.free) {
        return obj.free();
    }

}

export const loadFontSdf = () => {
    let filePath = PISYS.Env.get("sdfbinPath");
    if (filePath) {
        let info = Depend.getFile(filePath);
        if (info) {
            return loadRes(info).then((data) => {
                if (data instanceof ArrayBuffer) {
                    return new Uint8Array(data);
                } else {
                    return data;
                }
            })
        } else {
            return Promise.resolve(new Uint8Array());
        }
    } else {
        return Promise.resolve(new Uint8Array());
    }
}

